@@ -1,7 +1,8 @@
 /*
  * i-scream libstatgrab
  * http://www.i-scream.org
- * Copyright (C) 2000-2004 i-scream
+ * Copyright (C) 2000-2013 i-scream
+ * Copyright (C) 2010-2013 Jens Rehsack
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -17,7 +18,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  *
- * $Id: os_info.c,v 1.9 2004/11/06 18:58:23 tdb Exp $
+ * $Id$
  */
 
 #include <stdio.h>
@@ -25,25 +26,36 @@
 #include <stdlib.h>
 #include <unistd.h>
 
-int main(){
+#include "helpers.h"
+
+#ifndef lengthof
+#define lengthof(x) (sizeof(x)/sizeof((x)[0]))
+#endif
+
+static const char *host_states[] = {
+	"unknown", "physical host",
+	"virtual machine (full virtualized)",
+	"virtual machine (paravirtualized)",
+	"hardware virtualization"
+};
+
+int main(int argc, char **argv){
 
 	sg_host_info *general_stats;
 
+	/* Initialise helper - e.g. logging, if any */
+	sg_log_init("libstatgrab-examples", "SGEXAMPLES_LOG_PROPERTIES", argc ? argv[0] : NULL);
+
 	/* Initialise statgrab */
-	sg_init();
+	sg_init(1);
 
 	/* Drop setuid/setgid privileges. */
-	if (sg_drop_privileges() != 0) {
-		perror("Error. Failed to drop privileges");
-		return 1;
-	}
+	if (sg_drop_privileges() != SG_ERROR_NONE)
+		sg_die("Error. Failed to drop privileges", 1);
 
-	general_stats = sg_get_host_info();
-
-	if(general_stats == NULL){
-		fprintf(stderr, "Failed to get os stats\n");
-		exit(1);
-	}
+	general_stats = sg_get_host_info(NULL);
+	if(general_stats == NULL)
+		sg_die("Failed to get os stats", 1);
 
 	printf("OS name : %s\n", general_stats->os_name);
 	printf("OS release : %s\n", general_stats->os_release);
@@ -51,6 +63,12 @@ int main(){
 	printf("Hardware platform : %s\n", general_stats->platform);
 	printf("Machine nodename : %s\n", general_stats->hostname);
 	printf("Machine uptime : %lld\n", (long long)general_stats->uptime);
+	printf("Number of CPU's configured : %u\n", general_stats->maxcpus);
+	printf("Number of CPU's online : %u\n", general_stats->ncpus);
+	printf("Kernel bitwidth : %u\n", general_stats->bitwidth);
+	printf("Host state : %s\n", ((size_t)general_stats->host_state) > (lengthof(host_states) - 1)
+	                            ? "unexpected state (libstatgrab to new)"
+	                            : host_states[general_stats->host_state] );
 
 	exit(0);
 }
