@@ -1,7 +1,8 @@
 /*
  * i-scream libstatgrab
  * http://www.i-scream.org
- * Copyright (C) 2000-2004 i-scream
+ * Copyright (C) 2000-2013 i-scream
+ * Copyright (C) 2010-2013 Jens Rehsack
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -17,7 +18,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  *
- * $Id: process_stats.c,v 1.10 2004/04/06 14:52:56 tdb Exp $
+ * $Id$
  */
 
 #include <stdio.h>
@@ -25,6 +26,10 @@
 #include <stdlib.h>
 #include <unistd.h>
 
+#include "helpers.h"
+
+static int quit;
+
 int main(int argc, char **argv){
 
 	extern char *optarg;
@@ -41,22 +46,31 @@ int main(int argc, char **argv){
 		}
 	}
 
+	/* Initialise helper - e.g. logging, if any */
+	sg_log_init("libstatgrab-examples", "SGEXAMPLES_LOG_PROPERTIES", argc ? argv[0] : NULL);
+
 	/* Initialise statgrab */
-	sg_init();
+	sg_init(1);
 
 	/* Drop setuid/setgid privileges. */
-	if (sg_drop_privileges() != 0) {
-		perror("Error. Failed to drop privileges");
-		return 1;
-	}
+	if (sg_drop_privileges() != SG_ERROR_NONE)
+		sg_die("Error. Failed to drop privileges", 1);
+
+	register_sig_flagger( SIGINT, &quit );
+
+	if( (process_stat = sg_get_process_count()) == NULL )
+		sg_die("Failed to get process list", 1);
 
 	while( (process_stat = sg_get_process_count()) != NULL){
-		printf("Running:%3d \t", process_stat->running);
-		printf("Sleeping:%5d \t",process_stat->sleeping);
-		printf("Stopped:%4d \t", process_stat->stopped);
-		printf("Zombie:%4d \t", process_stat->zombie);
-		printf("Total:%5d\n", process_stat->total);
-		sleep(delay);
+		int ch;
+		printf("Running:%3llu \t", process_stat->running);
+		printf("Sleeping:%5llu \t",process_stat->sleeping);
+		printf("Stopped:%4llu \t", process_stat->stopped);
+		printf("Zombie:%4llu \t", process_stat->zombie);
+		printf("Total:%5llu\n", process_stat->total);
+		ch = inp_wait(delay);
+		if( quit || (ch == 'q') )
+			break;
 	}
 
 	exit(0);
